{************************************************************************}
{*                                                                      *}
{*      file       : RemoveClipboardFormatting.dpr                      *}
{*                                                                      *}
{*      type       : Application                                        *}
{*                                                                      *}
{*      location   : \OwnProjects\RemoveClipboardFormatting             *}
{*                                                                      *}
{*      purpose    : Utility to remove rich text from clipboard, or     *}
{*                   converts drag-drop data to plain text              *}
{*                                                                      *}
{*      author     : Lennert Ploeger                                    *}
{*                                                                      *}
{*      date       : 20021203                                           *}
{*                                                                      *}
{*      portability: 32 bit delphi only (V2.0 up)                       *}
{*                                                                      *}
{*      notes      : None                                               *}
{*                                                                      *}
{************************************************************************}
{* Updates:
When            Who     What
20021203        lsp     Creation
20040209        lsp     Write target of .lnk file to the clipboard instead of
                        the .lnk file itself, removed needless
                        WM_SYSCOMMAND, SC_HOTKEY code
20040421        lsp     Added the case-insensitive NO_SHORTCUT option
20070710        lsp     Made vista compliant by issuing Ctrl+Esc first.
20090707        lsp     CF_UNICODETEXT instead of CF_TEXT, IID_IShellLinkW instead of
                        IID_IShellLinkA
*}
program RemoveClipboardFormatting;

uses
  SysUtils, Shlobj, ActiveX, ComObj, Windows, Shellapi;

{$R *.RES}

{$warn unsafe_type off}
{$warn unsafe_code off}

type
  PShortCut = ^TShortCut;
  TShortCut = record
    Arguments: string;
    ShowCmd: Integer;
    WorkingDirectory: string;
    IdList: PItemIdList;
    FullPathAndNameOfLinkFile: string;
    Target: string;
    Description: string;
    IconLocation: string;
    IconIndex: Integer;
    HotKey: Word;
    FindData : TWIN32FINDDATA;
  end;
  TUnicodePath     = array[0..MAX_PATH-1] of WideChar;

const
  HOTKEYF_CONTROL       = 2;
  HOTKEYF_ALT           = 4;

var
  MyNewShortcut: TShortCut;
  sBuf: string;

// Helper routine to locate the target the .lnk file is pointing to
procedure GetTargetOfLink(lpShellLinkInfoStruct : PShortCut);
var
  ShellLink   : IShellLink;
  PersistFile : IPersistFile;
  MyObject  : IUnknown;
begin
  MyObject := CreateComObject(CLSID_ShellLink);
  ShellLink := MyObject as IShellLink;
  PersistFile := MyObject as IPersistFile;
  PersistFile.Load(
    PWChar(
      WideString(lpShellLinkInfoStruct^.FullPathAndNameOfLinkFile)),
      0);
  SetLength(lpShellLinkInfoStruct^.Target, MAX_PATH);
  ShellLink.GetPath(
    PChar(lpShellLinkInfoStruct^.Target),
    MAX_PATH,
    lpShellLinkInfoStruct^.FindData,
    SLGP_UNCPRIORITY);
  SetLength(lpShellLinkInfoStruct^.Target,
            StrLen(PChar(lpShellLinkInfoStruct^.Target)));
end;

procedure CopyToClipboard(const TextToCopy: string);
var
  DataHandle :  THandle;
  FromPointer:  Pointer;
  ToPointer  :  Pointer;
  s          :  string;
begin
  s := TextToCopy;
  SetLength(s, Length(s)+1);
  s[Length(s)] := #00;  // Null-terminator

  DataHandle := GlobalAlloc(GMEM_DDESHARE OR GMEM_MOVEABLE,
                            Length(s)*SizeOf(Char));

  ToPointer   := GlobalLock(DataHandle);
  FromPointer := @TextToCopy[1];
  Move(FromPointer^, ToPointer^, Length(s)*SizeOf(Char));
  GlobalUnlock(DataHandle);

  OpenClipboard(0);
  EmptyClipboard;
  SetClipboardData(CF_UNICODETEXT, DataHandle);
  CloseClipboard
end;

function GetClipboardText: string;
var
  Data       :  THandle;
begin
  OpenClipboard(0);
  Data := GetClipboardData(CF_UNICODETEXT);
  try
    if Data <> 0 then
      Result := PChar(GlobalLock(Data))
    else
      Result := '';
  finally
    if Data <> 0 then GlobalUnlock(Data);
  end;
  CloseClipboard
end;

function DragDropData: string;
var
  i: integer;
  MyHandle: THandle;
  num: integer;
  aFileName: array[0..MAX_PATH-1] of char;
  ShellLinkInfoStruct: TShortCut;
begin
  Result := '';
  OpenClipboard(0);
  try
    MyHandle := GetClipboardData(CF_HDROP);
    num := DragQueryFile(MyHandle, $FFFFFFFF, aFileName, 0);
    if num < 1 then Exit;
    for i:=0 to num-1 do
    begin
      DragQueryFile(MyHandle, i, aFileName, MAX_PATH);

      // Try to resolve .lnk files
      ShellLinkInfoStruct.FullPathAndNameOfLinkFile := aFileName;
      GetTargetOfLink(@ShellLinkInfoStruct);
      if (ShellLinkInfoStruct.Target <> '') then
        StrPCopy(aFileName, ShellLinkInfoStruct.Target);

      if Length(Result)>0 then
        Result := Result + #13#10 + aFileName
      else
        Result := aFileName
    end;
  finally
    CloseClipboard;
  end;
end;

function PidlToPath(const IdList: PItemIdList): string;
begin
  SetLength(Result, MAX_PATH);
  if SHGetPathFromIdList(IdList, PChar(Result)) then
    SetLength(Result, StrLen(PChar(Result)))
  else
    Result := '';
end;

function CreateShortCutOnDesktop(const ShortCut: TShortCut; const FileName: string): HRESULT;
var
  ShellLinkIntf : IShellLink;
  PersistFile   : IPersistFile;
  Path          : string;
  WidePath      : WideString;
  Pidl          : PItemIDList;
begin
  Result := E_FAIL;
  if Succeeded(SHGetSpecialFolderLocation(0, CSIDL_DESKTOP, Pidl)) then
  begin
    Path := PidltoPath(Pidl);
    if Path <> '' then
    begin
      SetLength(Path, StrLen(PChar(Path)));

      Result := CoCreateInstance(CLSID_ShellLink, nil, CLSCTX_INPROC_SERVER,
        IID_IShellLinkW, ShellLinkIntf);
      if Succeeded(Result) then
      begin
        ShellLinkIntf.SetArguments(PChar(ShortCut.Arguments));
        ShellLinkIntf.SetShowCmd(ShortCut.ShowCmd);
        ShellLinkIntf.SetWorkingDirectory(PChar(ShortCut.WorkingDirectory));
        ShellLinkIntf.SetPath(PChar(ShortCut.Target));
        ShellLinkIntf.SetDescription(PChar(ShortCut.Description));
        ShellLinkIntf.SetHotkey(ShortCut.HotKey);
        ShellLinkIntf.SetIconLocation(PChar(ShortCut.IconLocation), ShortCut.IconIndex);

        PersistFile := ShellLinkIntf as IPersistFile;
        WidePath := IncludeTrailingPathDelimiter(Path) + FileName;
        // Only store file when not already present
        if not FileExists(WidePath) then
          Result := PersistFile.Save(PWideChar(WidePath), True);
      end
    end
  end;
end;

begin
  CoInitialize(nil);
  if not FindCmdLineSwitch('NO_SHORTCUT', ['-', '/'], True) then
  begin
    MyNewShortcut.Arguments := '';
    MyNewShortcut.ShowCmd := SW_NORMAL;
    MyNewShortcut.WorkingDirectory := ExtractFilePath(ParamStr(0));
    MyNewShortcut.IdList := nil;
    MyNewShortcut.Target := ParamStr(0);
    MyNewShortcut.Description := 'Paste Plain Text from Clipboard!';
    MyNewShortcut.IconLocation := '';
    MyNewShortcut.IconIndex := 0;
    MyNewShortcut.HotKey := (HOTKEYF_CONTROL shl 8) or (HOTKEYF_ALT shl 8) or Ord('V');

    CreateShortCutOnDeskTop(MyNewShortcut, 'RemoveClipboardFormatting.lnk');
  end;

  if IsClipboardFormatAvailable(CF_TEXT) then
  begin
    sBuf := GetClipboardText;
    CopyToClipboard(sBuf);
  end
  else
    CopyToClipboard(DragDropData);

  // Issue the keystrokes

  // Vista moves focus to the taskbar (Shell_TrayWnd) when another program is started.
  // In this case be sure to activate nearest active window first by Ctrl+Esc.
  sBuf := StringOfChar(' ', 256);
  GetClassName(GetForeGroundWindow(), PChar(sBuf), Length(sBuf));
  if Pos('Shell_TrayWnd', sBuf)>0 then
  begin
    keybd_event( VK_MENU, 0, 0, 0);
    keybd_event( VK_ESCAPE, 0, 0, 0);
    // Sleep(50);
    keybd_event( VK_MENU, 0, KEYEVENTF_KEYUP, 0); {}
    keybd_event( VK_ESCAPE, 0, KEYEVENTF_KEYUP, 0);

    Sleep(10);
  end;

  keybd_event( VK_CONTROL, 0, 0, 0);
  keybd_event( Ord('V'), 0, 0, 0);
  keybd_event( VK_CONTROL, 0, KEYEVENTF_KEYUP, 0);
  keybd_event( Ord('V'), 0, KEYEVENTF_KEYUP, 0);
  CoUninitialize;
end.
